from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class CaptureTest(unittest.TestCase):

    def test_multi_capture(self):
        result = match(
            value=[{
                "first-name": "Jane",
                "last-name": "Doe",
            }, {
                "first-name": "John",
                "last-name": "Doe",
            }],
            pattern=Each({"first-name": Capture(..., name="first_names")}),
            multimatch=True,
        )
        self.assertTrue(result)
        self.assertEqual(["Jane", "John"], result['first_names'])

    def test_last_capture(self):
        result = match(
            multimatch=False,
            value=[{
                "first-name": "Jane",
                "last-name": "Doe",
            }, {
                "first-name": "John",
                "last-name": "Doe",
            }],
            pattern=Each({"first-name": Capture(..., name="first_name")})
        )
        self.assertTrue(result)
        self.assertEqual("John", result['first_name'])

    def test_syntactic_sugar_rshift_operator(self):
        result = match(sample_k8s_response, {
            "containers": Each({
                "image": _ >> 'image',
                "name": _ >> 'name',
                "ports": Each({
                    "containerPort": _ >> 'port'
                }),
            })
        })
        self.assertTrue(result)
        self.assertEqual('k8s.gcr.io/metrics-server/metrics-server:v0.4.1', result['image'])
        self.assertEqual('metrics-server', result['name'])
        self.assertEqual(4443, result['port'])

    def test_syntactic_sugar_rmatmul_operator(self):
        # noinspection PyUnresolvedReferences
        result = match(sample_k8s_response, {
            "containers": Each({
                "image": 'image' @ _,
                "name": 'name' @ _,
                "ports": Each({
                    "containerPort": 'port' @ _
                }),
            })
        })
        self.assertTrue(result)
        self.assertEqual('k8s.gcr.io/metrics-server/metrics-server:v0.4.1', result['image'])
        self.assertEqual('metrics-server', result['name'])
        self.assertEqual(4443, result['port'])

    def test_syntactic_sugar_underscore(self):
        result = match(sample_k8s_response, {
            "containers": Each({
                "image": _ >> 'image',
                "name": _ >> 'name',
                "ports": Each({
                    "containerPort": _ >> 'port'
                }),
            })
        })
        self.assertTrue(result)
        self.assertEqual('k8s.gcr.io/metrics-server/metrics-server:v0.4.1', result['image'])
        self.assertEqual('metrics-server', result['name'])
        self.assertEqual(4443, result['port'])

    def test_regex_bind_groups(self):
        result = match("abcdef", Regex(r"(?P<one>[a-c]+)(?P<two>[a-z]+)", bind_groups=True))
        self.assertTrue(result)
        self.assertEqual(result['one'], 'abc')
        self.assertEqual(result['two'], 'def')
        self.assertEqual(result.one, 'abc')
        self.assertEqual(result.two, 'def')

    def test_nested_capture(self):
        result = match([1, 2, 3, 4], [1, Many(...) >> 'x' >> 'y', 4])
        self.assertTrue(result)
        self.assertEqual([2, 3], result.x)
        self.assertEqual([2, 3], result.y)


sample_k8s_response = {
    "containers": [
        {
            "args": [
                "--cert-dir=/tmp",
                "--secure-port=4443",
                "--kubelet-preferred-address-types=InternalIP,ExternalIP,Hostname",
                "--kubelet-use-node-status-port"
            ],
            "image": "k8s.gcr.io/metrics-server/metrics-server:v0.4.1",
            "imagePullPolicy": "IfNotPresent",
            "livenessProbe": {
                "failureThreshold": 3,
                "httpGet": {
                    "path": "/livez",
                    "port": "https",
                    "scheme": "HTTPS"
                },
                "periodSeconds": 10,
                "successThreshold": 1,
                "timeoutSeconds": 1
            },
            "name": "metrics-server",
            "ports": [
                {
                    "containerPort": 4443,
                    "name": "https",
                    "protocol": "TCP"
                }
            ],
            "readinessProbe": {
                "failureThreshold": 3,
                "httpGet": {
                    "path": "/readyz",
                    "port": "https",
                    "scheme": "HTTPS"
                },
                "periodSeconds": 10,
                "successThreshold": 1,
                "timeoutSeconds": 1
            },
            "resources": {},
            "securityContext": {
                "readOnlyRootFilesystem": True,
                "runAsNonRoot": True,
                "runAsUser": 1000
            },
            "terminationMessagePath": "/dev/termination-log",
            "terminationMessagePolicy": "File",
            "volumeMounts": [
                {
                    "mountPath": "/tmp",
                    "name": "tmp-dir"
                },
                {
                    "mountPath": "/var/run/secrets/kubernetes.io/serviceaccount",
                    "name": "metrics-server-token-2j86f",
                    "readOnly": True
                }
            ]
        }
    ],
    "dnsPolicy": "ClusterFirst",
    "enableServiceLinks": True,
    "nodeName": "ip-10-2-169-118.ec2.internal",
    "nodeSelector": {
        "kubernetes.io/os": "linux"
    },
    "priority": 2000000000,
    "priorityClassName": "system-cluster-critical",
    "restartPolicy": "Always",
    "schedulerName": "default-scheduler",
    "securityContext": {},
    "serviceAccount": "metrics-server",
    "serviceAccountName": "metrics-server",
    "terminationGracePeriodSeconds": 30,
    "tolerations": [
        {
            "effect": "NoExecute",
            "key": "node.kubernetes.io/not-ready",
            "operator": "Exists",
            "tolerationSeconds": 300
        },
        {
            "effect": "NoExecute",
            "key": "node.kubernetes.io/unreachable",
            "operator": "Exists",
            "tolerationSeconds": 300
        }
    ],
    "volumes": [
        {
            "emptyDir": {},
            "name": "tmp-dir"
        },
        {
            "name": "metrics-server-token-2j86f",
            "secret": {
                "defaultMode": 420,
                "secretName": "metrics-server-token-2j86f"
            }
        }
    ]
}



# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import subprocess
import sys
from itertools import cycle
import threading
import os
import curses
import re
import warnings
from contextlib import suppress
import time
from pathlib import Path
from traceback import print_exc
from typing import List, AnyStr, Union

# noinspection PyUnresolvedReferences
from pyrsistent import pvector, pmap, pset
# noinspection PyUnresolvedReferences
from Aspidites._vendor.contracts import contract, new_contract
from Aspidites._vendor.pyparsing import ParseException, ParseResults
# noinspection PyUnresolvedReferences
from cmath import inf
# noinspection PyUnresolvedReferences
from Aspidites import *
import Aspidites.parser.parser
from Aspidites.api import _format_locals

try:
    import readline
except ImportError:
    readline = None
histfile = Path('~/.woma_shell_history').expanduser()
histfile_size = 1000
START_PROMPT = '>>> '
CONTINUE_PROMPT = '... '


class Spinner:  # pragma: no cover
    busy = False
    delay = 0.25

    def __init__(self, delay=None, stdout=sys.stdout):
        self.spinner_generator = (i for i in cycle('|/-\\'))
        self.stdout = stdout
        if delay and float(delay): self.delay = delay

    def spinner_task(self):
        while self.busy:
            self.stdout.write(next(self.spinner_generator))
            self.stdout.flush()
            time.sleep(self.delay)
            self.stdout.write('\b')
            self.stdout.flush()

    def __enter__(self):
        self.busy = True
        threading.Thread(target=self.spinner_task).start()

    def __exit__(self, exception, value, tb):
        self.busy = False
        time.sleep(self.delay)
        if exception is not None:
            return False


single_arg_help = re.compile(r'(?:help[\(])(\w+)(?:[\)])')


class ReadEvalParse:  # pragma: no cover
    intro = "Welcome to the Woma Interactive Shell. Use the 'help()' or '?' command to see a list of commands.\nThis is experimental and mainly aims to help developers to sandbox " \
            "Woma without compilation."
    ruler = "┉"
    doc_leader = ""
    nohelp = "*** No help on %s"
    doc_header = "Documented commands (type help <topic>):"
    misc_header = "Miscellaneous help topics:"
    undoc_header = "Undocumented commands:"
    _globals = globals().copy()

    def __init__(self, stdout=None):
        if stdout is not None:
            self.stdout = stdout
        else:
            self.stdout = sys.stdout
        self.warn = lambda x: sys.stderr.write(x) and sys.stderr.flush()
        self.__locals__ = dict(locals(), **globals())

    def input(self):
        self.stdout.flush()

        line = input(START_PROMPT)
        if ';' in line:
            line = line.replace(';', '\n    ')
            self.stdout.write(line + '\n')
            self.stdout.flush()
        return line

    def get_names(self):
        # This method used to pull in base class attributes
        # at a time dir() didn't do it yet.
        return dir(self.__class__)

    def find_token(self, token: str, text: str) -> bool:
        return text.find(token) != -1

    def displayhook(self, text):
        if text is None:
            return
        try:
            self.stdout.write(str(text))
        except UnicodeEncodeError:
            bytes = text.encode(sys.stdout.encoding, 'backslashreplace')
            if hasattr(self.stdout, 'buffer'):
                self.stdout.buffer.write(bytes)
            else:
                text = bytes.decode(sys.stdout.encoding, 'strict')
                self.stdout.write(text)
        self.stdout.write("\n")

    def eval_exec(self, x: Union[List, AnyStr]):
        # noinspection PyBroadException
        if isinstance(x, ParseResults):
            x = x[0]
        # noinspection PyBroadException
        warnings.resetwarnings()
        try:
            out = eval(compile(x, filename='<inline code>', mode='eval'),
                       self.__locals__,
                       self.__locals__)
        except Exception:
            out = exec(compile(x, filename='<inline code>', mode='exec'), self.__locals__, self.__locals__)
            self.stdout.write('\n')
            # if out is not None:
            #     print(out)
        self.displayhook(out)

    def preloop(self):
        if readline and Path(histfile).exists():
            readline.read_history_file(histfile)

    def postloop(self):
        if readline:
            readline.set_history_length(histfile_size)
            readline.write_history_file(histfile)

    def loop(self) -> None:
        os.system('cls' if os.name == 'nt' else 'clear')
        self.stdout.write(self.intro + '\n')
        try:
            while True:
                self.postloop()
                self.preloop()
                try:
                    line = self.input()
                    if line == '':
                        continue
                    if line == 'exit()':
                        raise SystemExit
                    if self.find_token('?', line):
                        self.do_help(line.lstrip('? '))
                        continue
                    if self.find_token('help ', line):
                        self.do_help(line.split(' ')[1])
                        continue
                    if single_arg_help.match(line):
                        self.do_help(single_arg_help.search(line).group(1))
                        continue
                    if hasattr(self, 'do_' + line):
                        getattr(self, 'do_' + line)()
                        continue
                    if str(line).isidentifier():
                        self.eval_exec(line)
                        continue
                    try:
                        with Spinner():
                            p = Aspidites.parser.parser.parse_module(line)
                    except ParseException:
                        self.warn(f'Warning: Failed to parse "{line}" as Woma.\n'
                              f'Remember that Woma does not allow literal evaluation, try assigning to a variable.\n'
                              f'Falling back to python with suppressed exceptions.\n')
                        with suppress(Exception):
                            self.eval_exec(line)
                        continue
                    else:
                        self.eval_exec(p)
                        continue

                except Exception as e:
                    self.stdout.write(f"Error: {e}\n")
                    continue
        except KeyboardInterrupt as e:
            self.do_exit()

    def do_exit(self, arg=None):
        """Exit the woma interactive interpreter."""
        self.stdout.write("\nExiting...")
        raise SystemExit

    def do_copyright(self):
        """Copyright Ross J. Duff 2021 licensed under the GNU Public License v3."""
        pass

    def do_locals(self, arg=None):
        """Print local variables"""
        hidden = ['self', 'stdout', 'ReadEvalParse', '__warningregistry__']
        visible = dict()
        for k, v in self.__locals__.items():
            if k in self._globals.keys() or k in hidden:
                continue
            else:
                visible[k] = v

        self.stdout.write(_format_locals(visible).decode('UTF-8'))

    def do_flush(self):
        """forcibly flush stdout"""
        self.stdout.flush()

    def do_help(self, arg=None):
        """List available commands with "help" or detailed help with "help cmd"."""
        if arg:
            # XXX check arg syntax
            try:
                func = getattr(self, 'help_' + arg)
            except AttributeError:
                try:
                    doc = getattr(self, 'do_' + arg).__doc__
                    if doc:
                        self.stdout.write("%s\n" % str(doc))
                        return
                except AttributeError:
                    pass
                self.stdout.write("%s\n" % str(self.nohelp % (arg,)))
                return
            func()
        else:
            names = self.get_names()
            cmds_doc = []
            cmds_undoc = []
            help = {}
            for name in names:
                if name[:5] == 'help_':
                    help[name[5:]] = 1
            names.sort()
            # There can be duplicates if routines overridden
            prevname = ''
            for name in names:
                if name[:3] == 'do_':
                    if name == prevname:
                        continue
                    prevname = name
                    cmd = name[3:]
                    if cmd in help:
                        cmds_doc.append(cmd)
                        del help[cmd]
                    elif getattr(self, name).__doc__:
                        cmds_doc.append(cmd)
                    else:
                        cmds_undoc.append(cmd)
            self.stdout.write("%s\n" % str(self.doc_leader))
            self.print_topics(self.doc_header, cmds_doc, 15, 80)
            self.print_topics(self.misc_header, list(help.keys()), 15, 80)
            self.print_topics(self.undoc_header, cmds_undoc, 15, 80)

    def print_topics(self, header, cmds, cmdlen, maxcol):
        if cmds:
            self.stdout.write("%s\n" % str(header))
            if self.ruler:
                self.stdout.write("╭%s╮\n" % str(self.ruler * len(header)))
            self.columnize(cmds, maxcol - 1)
            self.stdout.write("\n")

    def columnize(self, list, displaywidth=80):
        """Display a list of strings as a compact set of columns.

        Each column is only as wide as necessary.
        Columns are separated by two spaces (one was not legible enough).
        """
        if not list:
            self.stdout.write("<empty>\n")
            return

        nonstrings = [i for i in range(len(list))
                      if not isinstance(list[i], str)]
        if nonstrings:
            raise TypeError("list[i] not a string for i in %s"
                            % ", ".join(map(str, nonstrings)))
        size = len(list)
        if size == 1:
            self.stdout.write(' %s\n' % str(list[0]))
            return
        # Try every row count from 1 upwards
        for nrows in range(1, len(list)):
            ncols = (size + nrows - 1) // nrows
            colwidths = []
            totwidth = -2
            for col in range(ncols):
                colwidth = 0
                for row in range(nrows):
                    i = row + nrows * col
                    if i >= size:
                        break
                    x = list[i]
                    colwidth = max(colwidth, len(x))
                colwidths.append(colwidth)
                totwidth += colwidth + 2
                if totwidth > displaywidth:
                    break
            if totwidth <= displaywidth:
                break
        else:
            nrows = len(list)
            ncols = 1
            colwidths = [0]
        for row in range(nrows):
            texts = []
            for col in range(ncols):
                i = row + nrows * col
                if i >= size:
                    x = ""
                else:
                    x = list[i]
                texts.append(x)
            while texts and not texts[-1]:
                del texts[-1]
            for col in range(len(texts)):
                texts[col] = texts[col].ljust(colwidths[col])
            self.stdout.write(" %s\n" % str("  ".join(texts)))


if __name__ == "__main__":
    rep = ReadEvalParse()
    rep.loop()



""" Pyparsing parser for BibTeX files

A standalone parser using pyparsing.

pyparsing has a simple and expressive syntax so the grammar is easy to read and
write.

Submitted by Matthew Brett, 2010

Simplified BSD license
"""

from ....._vendor.pyparsing import (Regex, Suppress, ZeroOrMore, Group, Optional, Forward,
                       SkipTo, CaselessLiteral, Dict)


class Macro(object):
    """ Class to encapsulate undefined macro references """
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return 'Macro("%s")' % self.name
    def __eq__(self, other):
        return self.name == other.name
    def __ne__(self, other):
        return self.name != other.name


# Character literals
LCURLY,RCURLY,LPAREN,RPAREN,QUOTE,COMMA,AT,EQUALS,HASH = map(Suppress,'{}()",@=#')


def bracketed(expr):
    """ Return matcher for `expr` between curly brackets or parentheses """
    return (LPAREN + expr + RPAREN) | (LCURLY + expr + RCURLY)


# Define parser components for strings (the hard bit)
chars_no_curly = Regex(r"[^{}]+")
chars_no_curly.leaveWhitespace()
chars_no_quotecurly = Regex(r'[^"{}]+')
chars_no_quotecurly.leaveWhitespace()
# Curly string is some stuff without curlies, or nested curly sequences
curly_string = Forward()
curly_item = Group(curly_string) | chars_no_curly
curly_string << LCURLY + ZeroOrMore(curly_item) + RCURLY
# quoted string is either just stuff within quotes, or stuff within quotes, within
# which there is nested curliness
quoted_item = Group(curly_string) | chars_no_quotecurly
quoted_string = QUOTE + ZeroOrMore(quoted_item) + QUOTE

# Numbers can just be numbers. Only integers though.
number = Regex('[0-9]+')

# Basis characters (by exclusion) for variable / field names.  The following
# list of characters is from the btparse documentation
any_name = Regex('[^\\s"#%\'(),={}]+')

# btparse says, and the test bibs show by experiment, that macro and field names
# cannot start with a digit.  In fact entry type names cannot start with a digit
# either (see tests/bibs). Cite keys can start with a digit
not_digname = Regex('[^\\d\\s"#%\'(),={}][^\\s"#%\'(),={}]*')

# Comment comments out to end of line
comment = (AT + CaselessLiteral('comment') +
           Regex(r"[\s{(].*").leaveWhitespace())

# The name types with their digiteyness
not_dig_lower = not_digname.copy().setParseAction(lambda t: t[0].lower())
macro_def = not_dig_lower.copy()
macro_ref = not_dig_lower.copy().setParseAction(lambda t : Macro(t[0].lower()))
field_name = not_dig_lower.copy()
# Spaces in names mean they cannot clash with field names
entry_type = not_dig_lower('entry_type')
cite_key = any_name('cite_key')
# Number has to be before macro name
string = (number | macro_ref | quoted_string | curly_string)

# There can be hash concatenation
field_value = string + ZeroOrMore(HASH + string)
field_def = Group(field_name + EQUALS + field_value)
entry_contents = Dict(ZeroOrMore(field_def + COMMA) + Optional(field_def))

# Entry is surrounded either by parentheses or curlies
entry = (AT + entry_type + bracketed(cite_key + COMMA + entry_contents))

# Preamble is a macro-like thing with no name
preamble = AT + CaselessLiteral('preamble') + bracketed(field_value)

# Macros (aka strings)
macro_contents = macro_def + EQUALS + field_value
macro = AT + CaselessLiteral('string') + bracketed(macro_contents)

# Implicit comments
icomment = SkipTo('@').setParseAction(lambda t : t.insert(0, 'icomment'))

# entries are last in the list (other than the fallback) because they have
# arbitrary start patterns that would match comments, preamble or macro
definitions = Group(comment |
                    preamble |
                    macro |
                    entry |
                    icomment)

# Start symbol
bibfile = ZeroOrMore(definitions)


def parse_str(str):
    return bibfile.parseString(str)


if __name__ == '__main__':
    # Run basic test
    txt = """
Some introductory text
(implicit comment)

@ARTICLE{Authors2011,
  author = {First Author and Second Author and Third Author},
  title = {An article about {S}omething},
  journal = "Journal of Articles",
  year = {2011},
  volume = {16},
  pages = {1140--1141},
  number = {2}
}
"""
    print('\n\n'.join(defn.dump() for defn in parse_str(txt)))



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class TryMatchTest(unittest.TestCase):

    def test_try_match(self):
        result = None
        id_ = None
        try:
            match({
                "id": "h23ksad8",
                "name": "ObjectName"
            })
        except Case({"id": Capture(Regex("[A-Z0-9]+"), name='id')}):
            result = 'upper'
        except Case({"id": Capture(Regex("[a-z0-9]+"), name='id')}) as match_result:
            result = 'lower'
            id_ = match_result['id']
        finally:
            self.assertEqual('lower', result)
            self.assertEqual('h23ksad8', id_)

    def test_try_match_default(self):
        result = None
        try:
            match({
                "id": "h23ksad8",
                "name": "ObjectName"
            })
        except Case({"id": Regex("[A-Z0-9]+")}):
            result = 'upper'
        except Default:
            result = 'unknown'
        finally:
            self.assertEqual(result, 'unknown')



from __future__ import annotations

import unittest

# noinspection PyProtectedMember
import Aspidites._vendor.apm._util as util
from Aspidites._vendor.apm import *


# noinspection PyUnusedLocal
def f(a: int, b: float, c: str, d: str) -> str:
    pass


# noinspection PyUnusedLocal
def g(a: int, *, b: str) -> float:
    pass


# noinspection PyUnusedLocal
def h(a: int, *xs: float):
    pass


class UtilTests(unittest.TestCase):

    def test_get_arg_types(self):
        arg_types = util.get_arg_types(f)
        self.assertTrue(match(
            arg_types,
            [int, float, Remaining(str)]
        ))

    def test_get_return_type(self):
        self.assertEqual(str, util.get_return_type(f))

    def test_get_return_type_none(self):
        self.assertEqual(None, util.get_return_type(lambda: None))

    def test_invoke(self):
        args = {
            "b": 3,
            "c": 4,
        }
        result = util.invoke(lambda a, b: (a, b), args)
        self.assertEqual((None, 3), result)

    def test_get_kwarg_types(self):
        self.assertEqual({"a": int, "b": float, "c": str, "d": str}, util.get_kwarg_types(f))
        self.assertEqual({"a": int, "b": str}, util.get_kwarg_types(g))

    def test_vararg_types(self):
        self.assertEqual({"a": int}, util.get_kwarg_types(h))
        self.assertEqual([int, float], util.get_arg_types(h))



from __future__ import annotations

import re
import unittest
from dataclasses import dataclass

from Aspidites._vendor.apm import *


class TerseStyleTest(unittest.TestCase):

    def test_fibonacci(self):
        def fibonacci(n):
            return match(n,
                         1, 1,
                         2, 1,
                         _, lambda x: fibonacci(x - 1) + fibonacci(x - 2)
                         )

        self.assertEqual(1, fibonacci(1))
        self.assertEqual(1, fibonacci(2))
        self.assertEqual(2, fibonacci(3))
        self.assertEqual(3, fibonacci(4))
        self.assertEqual(5, fibonacci(5))
        self.assertEqual(8, fibonacci(6))

    def test_underscore_as_key(self):
        pet = {'type': 'dog', 'details': {'age': 3}}

        self.assertEqual(
            3,
            match(
                pet,
                {'details': {'age': _}}, lambda age: age
            )
        )
        self.assertEqual(
            ('details', 3),
            match(
                pet,
                {_: {'age': _}}, lambda a, b: (a, b)
            )
        )
        self.assertEqual(
            ('details', 3),
            match(
                pet,
                {_ >> 'key': {'age': _}}, lambda a, b: (a, b)
            )
        )

    def test_match_class_hierarchies(self):
        class Animal:
            pass

        class Hippo(Animal):
            pass

        class Zebra(Animal):
            pass

        class Horse(Animal):
            pass

        def what_is(x):
            return match(x,
                         Hippo, 'hippo',
                         Zebra, 'zebra',
                         Animal, 'some other animal',
                         _, 'not at all an animal',
                         )

        self.assertEqual('zebra', what_is(Zebra()))
        self.assertEqual('hippo', what_is(Hippo()))
        self.assertEqual('some other animal', what_is(Horse()))
        self.assertEqual('some other animal', what_is(Animal()))
        self.assertEqual('not at all an animal', what_is(42))

    def test_dataclasses(self):
        @dataclass
        class Pet:
            name: str
            age: int

        pet = Pet('rover', 7)

        self.assertEqual(7, match(pet, Pet('rover', _), lambda age: age))
        self.assertEqual('rover', match(pet, Pet(_, 7), lambda name: name))
        self.assertEqual(('rover', 7), match(pet, Pet(_, _), lambda name, age: (name, age)))

    def test_so_many_things(self):
        expectations = [
            (3, "this matches the number 3"),
            (4, "matches any integer"),
            (("k", 2), "a tuple (a, b) you can use in a function"),
            ([1, 2, 7], "any list of 3 elements that begins with [1, 2]"),
            ({'x': 3}, "any dict with a key 'x' and any value associated"),
            ({'x': 4, 'y': 4}, "any dict with a key 'x' and any value associated"),
            ("xyz", "anything else"),
        ]

        for val, expected in expectations:
            result = match(val,
                           3, "this matches the number 3",
                           int, "matches any integer",
                           (str, int), lambda a, b: "a tuple (a, b) you can use in a function",
                           [1, 2, _], "any list of 3 elements that begins with [1, 2]",
                           {'x': _}, "any dict with a key 'x' and any value associated",
                           _, "anything else"
                           )
            self.assertEqual(expected, result)

    def test_regex(self):
        def what_is(pet):
            return match(pet,
                         re.compile(r'(\w+)-(\w+)-cat$'), lambda name, my: 'cat ' + name,
                         re.compile(r'(\w+)-(\w+)-dog$'), lambda name, my: 'dog ' + name,
                         _, "something else"
                         )

        self.assertEqual('dog fuffy', what_is('fuffy-my-dog'))
        self.assertEqual('dog puffy', what_is('puffy-her-dog'))
        self.assertEqual('cat carla', what_is('carla-your-cat'))
        self.assertEqual('something else', what_is('roger-my-hamster'))

    def test_lambda_with_too_little_parameters(self):
        self.assertEqual("anything", match(1, _, lambda: "anything"))

    def test_lambda_with_too_many_parameters(self):
        self.assertEqual("anything(1, None, None)", match(1, _, lambda a, b, c: f"anything({a}, {b}, {c})"))

    def test_default_value(self):
        self.assertEqual("fallback", match(1, 2, lambda: "something", "fallback"))

    def test_default_action(self):
        self.assertEqual("fallback", match(1, 2, lambda: "something", lambda: "fallback"))



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from abc import ABCMeta
from types import FunctionType
import traceback


__all__ = ['ContractsMeta']


def is_function_or_static(f):
    is_normal_function: bool = isinstance(f, FunctionType)
    is_staticmethod: bool = isinstance(f, staticmethod)
    is_classmethod: bool = isinstance(f, classmethod)
    return is_normal_function or is_staticmethod or is_classmethod


class ContractsMeta(ABCMeta):
    """
        This metaclass lets the subclasses inherit the specifications.
        Very useful for abstract commands.

    """
    # TODO check that the contracts are a subtype of ContractsMeta
    def __init__(cls, clsname, bases, clsdict):  # @UnusedVariable @NoSelf
        ABCMeta.__init__(cls, clsname, bases, clsdict)

        for k, f in clsdict.items():
            is_normal_function = isinstance(f, FunctionType)
            is_staticmethod = isinstance(f, staticmethod)
            is_classmethod = isinstance(f, classmethod)

            if not (is_normal_function or is_staticmethod or is_classmethod):
                # print('skipping %s:%s, %s' % (clsname, k, f))
                continue
            if k == '__init__':
                continue

            # this_function = '%s:%s()' % (clsname, k)  # @UnusedVariable
            # print('considering %s' % this_function)

            superclasses = cls.mro()[1:]
            for b in superclasses:
                if k in b.__dict__:
                    f0 = b.__dict__[k]
                    if is_function_or_static(f0):
                        if isinstance(f0, classmethod):
                            # print('found ancestor classmethod')
                            pass
                        elif isinstance(f0, staticmethod):
                            # print('found ancestor staticmethod')
                            pass
                        else:
                            assert isinstance(f0, FunctionType)
                            if '__contracts__' in f0.__dict__:
                                spec = f0.__contracts__
                                # msg = 'inherit contracts for %s:%s() from %s' % (clsname, k, b.__name__)
                                # print(msg)
                                from Aspidites._vendor.contracts import ContractException
                                try:
                                    from .main import contracts_decorate
                                    f1 = contracts_decorate(f, **spec)
                                except ContractException as e:
                                    from .utils import indent
                                    msg = 'Error while applying ContractsMeta magic.\n'
                                    msg += '  subclass:  %s\n' % clsname
                                    msg += '      base:  %s\n' % b.__name__
                                    msg += '  function:  %s()\n' % k
                                    msg += 'Exception:\n'
                                    msg += indent(traceback.format_exc(), '| ') + '\n'
                                    msg += '(most likely parameters names are different?)'
                                    raise ContractException(msg)
                                setattr(cls, k, f1)
                                break
                    else:
                        pass
                else:
                    # print(' X not found in %s' % b.__name__)
                    pass


            else:
                pass
                # print(' -> No inheritance for %s' % this_function)


# This function is taken from six.
# https://bitbucket.org/gutworth/six
# six's permissive license allows us to include the following snippet.
def with_metaclass(meta, *bases):
    """Create a base class with a metaclass.

    This allows metaclasses with both python 2 and 3, which have different
    syntaxes for this functionality. Python 3's metaclass syntax is a
    SyntaxError in python 2.
    """
    # This requires a bit of explanation: the basic idea is to make a dummy
    # metaclass for one level of class instantiation that replaces itself with
    # the actual metaclass.
    class metaclass(meta):
        def __new__(cls, name, this_bases, d):
            return meta(name, bases, d)
    return type.__new__(metaclass, 'temporary_class', (), {})




from . import good, fail, syntax_fail


class A(object):
    a = 1
    b = 2

tc_a = A()


class B(object):
    a = 2

tc_b = B()



syntax_fail('attr')  # need at least some attribute

good('attr(a:*)', tc_a)
good('attr(a:int)', tc_a)
good('attr(b:int)', tc_a)
good('attr(b:>1)', tc_a)
good('attr(b:int,>1)', tc_a)
fail('attr(b:int,<=1)', tc_a)

good('attr(a:*)', tc_b)
fail('attr(b:*)', tc_b)

good('attr(a:int;b:int)', tc_a)
good('attr(a:int;b:int)', tc_a)
good('attr(a:int;b:int,>1)', tc_a)
fail('attr(a:int;b:int,<=1)', tc_a)



#
# decaf_parser.py
#
# Rudimentary parser for decaf language, used in Stanford University CS143
# (https://web.stanford.edu/class/archive/cs/cs143/cs143.1128/handouts/030%20Decaf%20Specification.pdf)
#
# To convert this parser into one that gives more of an AST, change all the Group wrappers to add parse
# actions that will result in ASTNode classes, or statement-specific subclasses.
#
# Copyright 2018, Paul McGuire
#
"""
    Program ::= Decl+
    Decl ::= VariableDecl | FunctionDecl  | ClassDecl | InterfaceDecl
    VariableDecl ::= Variable ; 
    Variable ::= Type ident 
    Type ::= int | double | bool | string | ident | Type [] 
    FunctionDecl ::= Type ident ( Formals ) StmtBlock | void ident ( Formals ) StmtBlock 
    Formals ::= Variable+, |  e 
    ClassDecl ::= class ident <extends ident>  <implements ident + ,>  { Field* } 
    Field ::= VariableDecl | FunctionDecl 
    InterfaceDecl ::= interface ident { Prototype* } 
    Prototype ::= Type ident ( Formals ) ; | void ident ( Formals ) ; 
    StmtBlock ::= { VariableDecl*  Stmt* } 
    Stmt ::=  <Expr> ; | IfStmt  | WhileStmt |  ForStmt | BreakStmt   | ReturnStmt  | PrintStmt  | StmtBlock 
    IfStmt ::= if ( Expr ) Stmt <else Stmt> 
    WhileStmt ::= while ( Expr ) Stmt 
    ForStmt ::= for ( <Expr> ; Expr ; <Expr> ) Stmt 
    ReturnStmt ::= return <Expr> ; 
    BreakStmt ::= break ; 
    PrintStmt ::= Print ( Expr+, ) ; 
    Expr ::= LValue = Expr | Constant | LValue | this | Call
            | ( Expr ) 
            | Expr + Expr | Expr - Expr | Expr * Expr | Expr / Expr |  Expr % Expr | - Expr 
            | Expr < Expr | Expr <= Expr | Expr > Expr | Expr >= Expr | Expr == Expr | Expr != Expr 
            | Expr && Expr | Expr || Expr | ! Expr 
            | ReadInteger ( ) | ReadLine ( ) | new ident | NewArray ( Expr , Typev) 
    LValue ::= ident |  Expr  . ident | Expr [ Expr ] 
    Call ::= ident  ( Actuals ) |  Expr  .  ident  ( Actuals ) 
    Actuals ::=  Expr+, | e 
    Constant ::= intConstant | doubleConstant | boolConstant |  stringConstant | null
"""
from ....._vendor import pyparsing as pp
from ....._vendor.pyparsing import pyparsing_common as ppc
pp.ParserElement.enablePackrat()

# keywords
keywords = (VOID, INT, DOUBLE, BOOL, STRING, CLASS, INTERFACE, NULL, THIS, EXTENDS, IMPLEMENTS, FOR, WHILE,
            IF, ELSE, RETURN, BREAK, NEW, NEWARRAY, PRINT, READINTEGER, READLINE, TRUE, FALSE) = map(pp.Keyword,
            """void int double bool string class interface null this extends implements or while
               if else return break new NewArray Print ReadInteger ReadLine true false""".split())
keywords = pp.MatchFirst(list(keywords))

LPAR, RPAR, LBRACE, RBRACE, LBRACK, RBRACK, DOT, EQ, COMMA, SEMI = map(pp.Suppress, "(){}[].=,;")
hexConstant = pp.Regex(r"0[xX][0-9a-fA-F]+").addParseAction(lambda t: int(t[0][2:], 16))
intConstant = hexConstant | ppc.integer
doubleConstant = ppc.real
boolConstant = TRUE | FALSE
stringConstant = pp.dblQuotedString
null = NULL
constant = doubleConstant | boolConstant | intConstant | stringConstant | null
ident = ~keywords + pp.Word(pp.alphas, pp.alphanums+'_')
type_ = pp.Group((INT | DOUBLE | BOOL | STRING | ident) + pp.ZeroOrMore("[]"))

variable = type_ + ident
variable_decl = variable + SEMI

expr = pp.Forward()
expr_parens = pp.Group(LPAR + expr + RPAR)
actuals = pp.Optional(pp.delimitedList(expr))
call = pp.Group(ident("call_ident") + LPAR + actuals("call_args") + RPAR
                | (expr_parens + pp.ZeroOrMore(DOT + ident))("call_ident_expr") + LPAR + actuals("call_args") + RPAR)
lvalue = ((ident | expr_parens) 
           + pp.ZeroOrMore(DOT + (ident | expr_parens)) 
           + pp.ZeroOrMore(LBRACK + expr + RBRACK))
assignment = pp.Group(lvalue("lhs") + EQ + expr("rhs"))
read_integer = pp.Group(READINTEGER + LPAR + RPAR)
read_line = pp.Group(READLINE + LPAR + RPAR)
new_statement = pp.Group(NEW + ident)
new_array = pp.Group(NEWARRAY + LPAR + expr + COMMA + type_ + RPAR)
rvalue = constant | call | read_integer | read_line | new_statement | new_array | ident
arith_expr = pp.infixNotation(rvalue,
    [
    ('-', 1, pp.opAssoc.RIGHT,),
    (pp.oneOf("* / %"), 2, pp.opAssoc.LEFT,),
    (pp.oneOf("+ -"), 2, pp.opAssoc.LEFT,),
    ])
comparison_expr = pp.infixNotation(arith_expr, 
    [
    ('!', 1, pp.opAssoc.RIGHT,),
    (pp.oneOf("< > <= >="), 2, pp.opAssoc.LEFT,),
    (pp.oneOf("== !="), 2, pp.opAssoc.LEFT,),
    (pp.oneOf("&&"), 2, pp.opAssoc.LEFT,),
    (pp.oneOf("||"), 2, pp.opAssoc.LEFT,),
    ])
expr <<= (assignment
          | call
          | THIS
          | comparison_expr
          | arith_expr
          | lvalue
          | constant
          | read_integer
          | read_line
          | new_statement
          | new_array
          )

stmt = pp.Forward()
print_stmt = pp.Group(PRINT("statement") + LPAR + pp.Group(pp.Optional(pp.delimitedList(expr)))("args") + RPAR + SEMI)
break_stmt = pp.Group(BREAK("statement") + SEMI)
return_stmt = pp.Group(RETURN("statement") + expr + SEMI)
for_stmt = pp.Group(FOR("statement") + LPAR + pp.Optional(expr) + SEMI + expr + SEMI + pp.Optional(expr) + RPAR + stmt)
while_stmt = pp.Group(WHILE("statement") + LPAR + expr + RPAR + stmt)
if_stmt = pp.Group(IF("statement")
                   + LPAR + pp.Group(expr)("condition") + RPAR
                   + pp.Group(stmt)("then_statement")
                   + pp.Group(pp.Optional(ELSE + stmt))("else_statement"))
stmt_block = pp.Group(LBRACE + pp.ZeroOrMore(variable_decl) + pp.ZeroOrMore(stmt) + RBRACE)
stmt <<= (if_stmt
          | while_stmt
          | for_stmt
          | break_stmt
          | return_stmt
          | print_stmt
          | stmt_block
          | pp.Group(expr + SEMI)
          )

formals = pp.Optional(pp.delimitedList(variable))
prototype = pp.Group((type_ | VOID)("return_type")
                     + ident("function_name")
                     + LPAR + formals("args") + RPAR + SEMI)("prototype")
function_decl = pp.Group((type_ | VOID)("return_type") + ident("function_name")
                         + LPAR + formals("args") + RPAR
                         + stmt_block("body"))("function_decl")

interface_decl = pp.Group(INTERFACE + ident("interface_name")
                          + LBRACE + pp.ZeroOrMore(prototype)("prototypes") + RBRACE)("interface")
field = variable_decl | function_decl
class_decl = pp.Group(CLASS + ident("class_name")
              + pp.Optional(EXTENDS + ident)("extends")
              + pp.Optional(IMPLEMENTS + pp.delimitedList(ident))("implements")
              + LBRACE + pp.ZeroOrMore(field)("fields") + RBRACE)("class_decl")

decl = variable_decl | function_decl | class_decl | interface_decl | prototype
program = pp.OneOrMore(pp.Group(decl))
decaf_parser = program

stmt.runTests("""\
    sin(30);
    a = 1;
    b = 1 + 1;
    b = 1 != 2 && false;
    print("A");
    a.b = 100;
    a.b = 100.0;
    a[100] = b;
    a[0][0] = 2;
    a = 0x1234;
""")

test_program = """
    void getenv(string var);
    int main(string[] args) {
        if (a > 100) {
            Print(a, " is too big");
        } else if (a < 100) {
            Print(a, " is too small");
        } else {
            Print(a, "just right!");
        }
    }
"""

print(decaf_parser.parseString(test_program).dump())



# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=True, initializedcheck=False
# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from Aspidites._vendor.pyparsing import (
    Combine,
    Empty,
    FollowedBy,
    Forward,
    ZeroOrMore,
    OneOrMore,
    MatchFirst,
    Optional,
    ParseElementEnhance,
    ParseException,
    Regex,
    Char,
    Word,
    alphanums,
    alphas,
    col,
    conditionAsParseAction,
    delimitedList,
    infixNotation,
    matchOnlyAtCol,
    nums,
    opAssoc,
    quotedString,
    unicodeString,
)

from .._vendor.contracts.syntax import contract_expression
from ..parser.convert import *

_contract_expression = contract_expression.copy()
_contract_expression.setParseAction(lambda tks: f"'{''.join((str(t) for t in tks))}'")


class IndentedBlock(ParseElementEnhance):
    """
    Expression to match one or more expressions at a given indentation level.
    Useful for parsing text where structure is implied by indentation (like Python source code).
    """

    def __init__(self, expr, recursive=True):
        super().__init__(expr, savelist=True)
        self._recursive = recursive
        self.skipWhitespace = False

    def parseImpl(self, instring, loc, doActions=True):
        # see if self.expr matches at the current location - if not it will raise an exception
        # and no further work is necessary
        self.expr.parseImpl(instring, loc, doActions)

        indent_col = col(loc, instring)
        peer_parse_action = matchOnlyAtCol(indent_col)
        peer_expr = FollowedBy(self.expr).addParseAction(peer_parse_action)
        inner_expr = Empty() + peer_expr.suppress() + self.expr

        if self._recursive:
            indent_parse_action = conditionAsParseAction(
                lambda s, l, t, relative_to_col=indent_col: col(l, s) > relative_to_col
            )
            indent_expr = FollowedBy(self.expr).addParseAction(indent_parse_action)
            inner_expr += Optional(indent_expr + self)

        return OneOrMore(inner_expr).parseImpl(instring, loc, doActions)


quoted_str = quotedString().setParseAction(lambda t: t[0])
integer = Word(nums).setParseAction(cvt_int)
# TODO .5 not recognized as a float
real = Combine(Optional(Word(nums)) + "." + Word(nums))
complex_ = Combine(real | integer + "+" + real | integer + "j")
list_str = Forward()
list_str_evolver = Forward()
set_str = Forward()
set_str_evolver = Forward()
dict_str = Forward()
dict_str_evolver = Forward()
tuple_str = Forward()
slice_str = Forward()
simple_assign = Forward()
func_call = Forward()
suite = Forward()
rvalue = Forward()
stmt = Forward()
identifier = Word(alphas + "_", alphanums + "_")
operand = nullit | complex_ | real | bool_literal | integer | identifier | underscore
arith_expr = Combine(
    infixNotation(
        operand,
        [
            (factop, 1, opAssoc.LEFT),
            (exponp, 2, opAssoc.RIGHT),
            (signop, 1, opAssoc.RIGHT),  # Changing this to LEFT causes the entire parser to break
            (multop, 2, opAssoc.LEFT),
            (plusop, 2, opAssoc.LEFT),
            (bitwiseop, 2, opAssoc.LEFT),
        ],
        lpar=lit_lparen,
        rpar=lit_rparen,
    )
).setParseAction(cvt_arith_expr)
comp_expr = infixNotation(
    arith_expr,
    [
        (comparisonop, 2, opAssoc.LEFT),
    ],
).setParseAction(lambda t: ''.join(t[0]))

collection = quoted_str | list_str | set_str | tuple_str | dict_str

list_item = (  # Precedence important!!!
        slice_str
        | comp_expr  # Expressions
        | arith_expr
        | identifier
        | integer  # Literals
        | complex_
        | real
        | bool_literal
        | nullit
        | list_str_evolver
        | set_str_evolver
        | dict_str_evolver
        | collection
)

assignable = func_call | list_item

contract_define = identifier + imposes + _contract_expression
contract_define.setParseAction(cvt_contract_define)
contract_respect = respects + _contract_expression
contract_assign = identifier + assign_eq + assignable + contract_respect
contract_assign.setParseAction(cvt_contract_assign)

tuple_str <<= (lparen + Optional(delimitedList(list_item)) + Optional(comma) + rparen).setParseAction(cvt_tuple)
list_str <<= (lbrack + Optional(delimitedList(list_item) + Optional(comma)) + rbrack).setParseAction(cvt_list)
list_str_evolver <<= ((lbrack + Optional(delimitedList(list_item) + Optional(comma)) + rbrack).setParseAction(
    cvt_list) + noclosure.setParseAction(replaceWith('.evolver()'))).setParseAction(lambda t: ''.join(t))
set_str <<= (lbrace + Optional(delimitedList(list_item) + Optional(comma)) + rbrace).setParseAction(cvt_set)
set_str_evolver <<= ((lbrace + Optional(delimitedList(list_item) + Optional(comma)) + rbrace).setParseAction(
    cvt_set) + noclosure.setParseAction(replaceWith('.evolver()'))).setParseAction(lambda t: ''.join(t))
dict_entry = Group(list_item + colon + list_item)
dict_str <<= (lbrace + Optional(delimitedList(dict_entry) + Optional(comma)) + rbrace).setParseAction(cvt_dict)
dict_str_evolver <<= ((lbrace + Optional(delimitedList(dict_entry) + Optional(comma)) + rbrace).setParseAction(
    cvt_dict) + noclosure.setParseAction(replaceWith('.evolver()'))).setParseAction(lambda t: ''.join(t))
slice_str <<= identifier + lit_lbrack + (integer | identifier) + Optional(
    lit_colon + (integer | identifier)) + lit_rbrack
slice_str.setParseAction(lambda t: ''.join(str(i) for i in t))
def_args = Optional(delimitedList(contract_assign, delim=";")).setParseAction(lambda t: sep.join(t))
def_args = Group(lit_lparen + def_args + args_end).setParseAction(lambda t: "".join(*t))

bool_pragmas = Combine(
    pragma + oneOf(' '.join(available_bool_pragmas)) + lit_lparen + oneOf('True False') + lit_rparen
).setParseAction(cvt_pragma)

func_decl = Group(
    Optional(OneOrMore(bool_pragmas)) + MatchFirst(
        private_def_decl | c_def_decl) + identifier + def_args + _contract_expression
).setParseAction(lambda t: "\n@contract()\n" + "".join(*t) + lit_colon)
comment_line = (
    Combine(Regex(r"`(?:[^`\n\r\\]|(?:``)|(?:\\(?:[^x]|x[0-9a-fA-F]+)))*") + "`")
        .setParseAction(lambda t: t[0])
        .setParseAction(cvt_comment_line)
)

return_value = return_none + rvalue
yield_value = yield_none + rvalue
ret_stmt = Group(return_value).setParseAction(lambda t: "".join(*t))
yield_stmt = Group(yield_value).setParseAction(lambda t: "".join(*t))

func_def = Group(func_decl + suite).setParseAction(lambda t: nl_indent.join(t[0]) + "\n")
lambda_def = Combine(Group(lit_lparen + arith_expr | comp_expr + lit_rparen))
func_call <<= Group(identifier + lit_lparen + Optional(delimitedList(rvalue)) + lit_rparen).setParseAction(
    cvt_func_call)  # if len(t[0]) != 3 else t[0][0] + '()')
clos_call = Group(identifier + lit_lparen + Optional(delimitedList(rvalue)) + lit_rparen).setParseAction(
    cvt_clos_call) + Suppress(noclosure)
context_suite = Forward()
context_def = Group(context_stmt + func_call).setParseAction(lambda t: ''.join(*t) + lit_colon)

context_decl = Group(context_def + context_suite).setParseAction(
    lambda t: ''.join(*t))

# TODO (!): trigram only binds a single letter variable identifier
lit_ellipse = Literal("...").setParseAction(replaceWith('...'))
case_stmt = Group(rvalue + colon + func_call).setParseAction(lambda t: sep.join(t[0]))
match_suite = Group(IndentedBlock(OneOrMore(case_stmt))).setParseAction(lambda t: (sep.join(t.asList()[0])))
match_decl = Group(match_none + Char(alphas)).setParseAction(lambda t: t[0][1] + '=' + t[0][0] + lit_lparen + t[0][1])
match_def = Group(match_decl + match_suite).setParseAction(lambda t: sep.join(t[0]) + lit_rparen)

loop_suite = Forward()

ident_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + identifier).setParseAction(
    cvt_for_loop_decl)
ident_loop_def = Group(ident_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

list_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + list_str).setParseAction(
    cvt_for_loop_decl)
list_loop_def = Group(list_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

set_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + set_str).setParseAction(
    cvt_for_loop_decl)
set_loop_def = Group(set_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

tuple_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + tuple_str).setParseAction(
    cvt_for_loop_decl)
tuple_loop_def = Group(tuple_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

dict_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + dict_str).setParseAction(
    cvt_for_loop_decl)
dict_loop_def = Group(dict_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

string_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + quoted_str).setParseAction(
    cvt_for_loop_decl)
string_loop_def = Group(string_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

func_loop_decl = Group(identifier + Optional(lit_comma + identifier) + for_none + func_call).setParseAction(
    cvt_for_loop_decl)
func_loop_def = Group(func_loop_decl + loop_suite).setParseAction(lambda t: nl_indent.join(t[0]))

loop_suite <<= IndentedBlock(
    OneOrMore(pass_stmt
              | cont_stmt
              | break_stmt
              | func_call
              | match_def
              ))
# TODO context managers get eaten by the preceding code blocks
context_suite <<= IndentedBlock(OneOrMore(contract_assign
                                          | match_def
                                          | func_call)).setParseAction(
    lambda t: (nl_indent.join(t.asList())))

suite <<= IndentedBlock(
    OneOrMore(pass_stmt
              | ret_stmt
              | yield_stmt
              | func_call
              | match_def
              | contract_assign)).setParseAction(
    lambda t: (nl_indent.join(t.asList())))
rvalue <<= clos_call | func_call | list_item | lambda_def | lit_ellipse
simple_assign << Group(identifier + assign_eq + rvalue).setParseAction(lambda t: " ".join(t[0]))
stmt <<= (func_def
          | match_def
          | ident_loop_def
          | string_loop_def
          | list_loop_def
          | set_loop_def
          | tuple_loop_def
          | dict_loop_def
          | func_loop_def
          | contract_define
          | func_call
          | simple_assign)
module_body = OneOrMore(stmt) + Optional(
    struct_main + OneOrMore(stmt).setParseAction(lambda t: indent + nl_indent.join(t)))
module_body.ignore(comment_line)


def parse_module(module):
    return module_body.parseString(module, parseAll=True)


def parse_statement(s):
    return stmt.parseString(s)

# if __name__ == "__main__":
#     print(parse_module(open('../examples/math.wom').read()))



__version__ = '1.3.3'
__title__ = 'Aspidites'
__author__ = 'Ross J. Duff'
__license__ = "GPL"
__description__ = (
    f"""
    {__title__} v{__version__} is the reference Woma programming language compiler.\n
    Copyright (C) 2021  {__author__}\n
    This program comes with ABSOLUTELY NO WARRANTY;\n
    This is free software, and you are welcome to redistribute it\n
    under the conditions of the {__license__}v3 found in the LICENSE file.
    """
)
__mimetype__ = 'text/woma'
__all__ = [
    'SafeMod',
    'SafeDiv',
    'SafeExp',
    'SafeFloorDiv',
    'SafeFactorial',
    'SafeUnarySub',
    'SafeUnaryAdd',
    'Undefined',
    'Maybe',
    'Surely',
    'Warn',
    'parse_module',
    'parse_statement',
    'bordered'
    ]

from Aspidites.api import *
from Aspidites.parser import *
from Aspidites.monads import *
from Aspidites.math import *

Maybe.__doc__ = Maybe.__doc__


#
# lucene_grammar.py
#
# Copyright 2011, Paul McGuire
#
# implementation of Lucene grammar, as decribed
# at http://svn.apache.org/viewvc/lucene/dev/trunk/lucene/docs/queryparsersyntax.html
#

from ....._vendor import pyparsing as pp
from pyparsing import pyparsing_common as ppc
pp.ParserElement.enablePackrat()

COLON,LBRACK,RBRACK,LBRACE,RBRACE,TILDE,CARAT = map(pp.Literal,":[]{}~^")
LPAR,RPAR = map(pp.Suppress,"()")
and_, or_, not_, to_ = map(pp.CaselessKeyword, "AND OR NOT TO".split())
keyword = and_ | or_ | not_ | to_

expression = pp.Forward()

valid_word = pp.Regex(r'([a-zA-Z0-9*_+.-]|\\\\|\\([+\-!(){}\[\]^"~*?:]|\|\||&&))+').setName("word")
valid_word.setParseAction(
    lambda t : t[0].replace('\\\\',chr(127)).replace('\\','').replace(chr(127),'\\')
    )

string = pp.QuotedString('"')

required_modifier = pp.Literal("+")("required")
prohibit_modifier = pp.Literal("-")("prohibit")
integer = ppc.integer()
proximity_modifier = pp.Group(TILDE + integer("proximity"))
number = ppc.fnumber()
fuzzy_modifier = TILDE + pp.Optional(number, default=0.5)("fuzzy")

term = pp.Forward()
field_name = valid_word().setName("fieldname")
incl_range_search = pp.Group(LBRACK - term("lower") + to_ + term("upper") + RBRACK)
excl_range_search = pp.Group(LBRACE - term("lower") + to_ + term("upper") + RBRACE)
range_search = incl_range_search("incl_range") | excl_range_search("excl_range")
boost = (CARAT - number("boost"))

string_expr = pp.Group(string + proximity_modifier) | string
word_expr = pp.Group(valid_word + fuzzy_modifier) | valid_word
term << (pp.Optional(field_name("field") + COLON)
         + (word_expr | string_expr | range_search | pp.Group(LPAR + expression + RPAR))
         + pp.Optional(boost))
term.setParseAction(lambda t:[t] if 'field' in t or 'boost' in t else None)

expression << pp.infixNotation(term,
    [
    (required_modifier | prohibit_modifier, 1, pp.opAssoc.RIGHT),
    ((not_ | '!').setParseAction(lambda: "NOT"), 1, pp.opAssoc.RIGHT),
    ((and_ | '&&').setParseAction(lambda: "AND"), 2, pp.opAssoc.LEFT),
    (pp.Optional(or_ | '||').setParseAction(lambda: "OR"), 2, pp.opAssoc.LEFT),
    ])

if __name__ == '__main__':

    # test strings taken from grammar description doc, and TestQueryParser.java
    tests = r"""
        # Success tests
        a and b
        a and not b
        a and !b
        a && !b
        a&&!b
        name:a
        name:a and not title:b
        (a^100 c d f) and !z
        name:"blah de blah"
        title:(+return +"pink panther")
        title:"The Right Way" AND text:go
        title:"Do it right" AND right
        title:Do it right
        roam~
        roam~0.8
        "jakarta apache"~10
        mod_date:[20020101 TO 20030101]
        title:{Aida TO Carmen}
        jakarta apache
        jakarta^4 apache
        "jakarta apache"^4 "Apache Lucene"
        "jakarta apache" jakarta
        "jakarta apache" OR jakarta
        "jakarta apache" AND "Apache Lucene"
        +jakarta lucene
        "jakarta apache" NOT "Apache Lucene"
        "jakarta apache" -"Apache Lucene"
        (jakarta OR apache) AND website
        \(1+1\)\:2
        c\:\\windows
        (fieldX:xxxxx OR fieldy:xxxxxxxx)^2 AND (fieldx:the OR fieldy:foo)
        (fieldX:xxxxx fieldy:xxxxxxxx)^2 AND (fieldx:the fieldy:foo)
        (fieldX:xxxxx~0.5 fieldy:xxxxxxxx)^2 AND (fieldx:the fieldy:foo)
        +term -term term
        foo:term AND field:anotherTerm
        germ term^2.0
        (term)^2.0
        (foo OR bar) AND (baz OR boo)
        +(apple \"steve jobs\") -(foo bar baz)
        +title:(dog OR cat) -author:\"bob dole\"
        a AND b
        +a +b
        (a AND b)
        c OR (a AND b)
        c (+a +b)
        a AND NOT b
        +a -b
        a AND -b
        a AND !b
        a && b
        a && ! b
        a OR b
        a b
        a || b
        a OR !b
        a -b
        a OR ! b
        a OR -b
        a - b
        a + b
        a ! b
        +foo:term +anotherterm
        hello
        term^2.0
        (germ term)^2.0
        term^2
        +(foo bar) +(baz boo)
        ((a OR b) AND NOT c) OR d
        (+(a b) -c) d
        field
        a&&b
        .NET
        term
        germ
        3
        term 1.0 1 2
        term term1 term2
        term term term
        term*
        term*^2
        term*^2.0
        term~
        term~2.0
        term~0.7
        term~^3
        term~2.0^3.0
        term*germ
        term*germ^3
        term*germ^3.0
        term~1.1
        [A TO C]
        t*erm*
        *term*
        term term^3.0 term
        term stop^3.0 term
        term +stop term
        term -stop term
        drop AND (stop) AND roll
        +drop +roll
        term +(stop) term
        term -(stop) term
        drop AND stop AND roll
        term phrase term
        term (phrase1 phrase2) term
        term AND NOT phrase term
        +term -(phrase1 phrase2) term
        stop^3
        stop
        (stop)^3
        ((stop))^3
        (stop^3)
        ((stop)^3)
        (stop)
        ((stop))
        term +stop
        [ a TO z]
        [a TO z]
        [ a TO z ]
        { a TO z}
        {a TO z}
        { a TO z }
        { a TO z }^2.0
        {a TO z}^2.0
        [ a TO z] OR bar
        [a TO z] bar
        [ a TO z] AND bar
        +[a TO z] +bar
        ( bar blar { a TO z})
        bar blar {a TO z}
        gack ( bar blar { a TO z})
        gack (bar blar {a TO z})
        [* TO Z]
        [* TO z]
        [A TO *]
        [a TO *]
        [* TO *]
        [\* TO \*]
        \!blah
        \:blah
        blah
        \~blah
        \*blah
        a
        a-b:c
        a+b:c
        a\:b:c
        a\\b:c
        a:b-c
        a:b+c
        a:b\:c
        a:b\\c
        a:b-c*
        a:b+c*
        a:b\:c*
        a:b\\c*
        a:b-c~2.0
        a:b+c~2.0
        a:b\:c~
        a:b\\c~
        [a- TO a+]
        [ a\\ TO a\* ]
        c\:\\temp\\\~foo.txt
        abc
        XYZ
        (item:\\ item:ABCD\\)
        \*
        *
        \\
        \||
        \&&
        a\:b\:c
        a\\b\:c
        a\:b\\c
        a\:b\:c\*
        a\:b\\\\c\*
        a:b-c~
        a:b+c~
        a\:b\:c\~
        a\:b\\c\~
        +weltbank +worlbank
        +term +term +term
        term +term term
        term term +term
        term +term +term
        -term term term
        -term +term +term
        on
        on^1.0
        hello^2.0
        the^3
        the
        some phrase
        xunit~
        one two three
        A AND B OR C AND D
        +A +B +C +D
        foo:zoo*
        foo:zoo*^2
        zoo
        foo:*
        foo:*^2
        *:foo
        a:the OR a:foo
        a:woo OR a:the
        *:*
        (*:*)
        +*:* -*:*
        the wizard of ozzy
        """

    failtests = r"""
        # Failure tests

        # multiple ':'s in term
        field:term:with:colon some more terms

        # multiple '^'s in term
        (sub query)^5.0^2.0 plus more
        a:b:c
        a:b:c~
        a:b:c*
        a:b:c~2.0
        \+blah
        \-blah
        foo \|| bar
        foo \AND bar
        \a
        a\-b:c
        a\+b:c
        a\b:c
        a:b\-c
        a:b\+c
        a\-b\:c
        a\+b\:c
        a:b\c*
        a:b\-c~
        a:b\+c~
        a:b\c
        a:b\-c*
        a:b\+c*
        [ a\- TO a\+ ]
        [a\ TO a*]
        a\\\+b
        a\+b
        c:\temp\~foo.txt
        XY\
        a\u0062c
        a:b\c~2.0
        XY\u005a
        XY\u005A
        item:\ item:ABCD\
        \
        a\ or b
        a\:b\-c
        a\:b\+c
        a\:b\-c\*
        a\:b\+c\*
        a\:b\-c\~
        a\:b\+c\~
        a:b\c~
        [ a\ TO a* ]
        """

    success1, _ = expression.runTests(tests)
    success2, _ = expression.runTests(failtests, failureTests=True)

    print("All tests:", ("FAIL", "OK")[success1 and success2])

    if not (success1 and success2):
        import sys
        sys.exit(1)



from . import good, fail
import collections

# sequences
for obj in ([], tuple(), set(), frozenset(), collections.deque(), {}, 'ciao',):
    good('collection[*]', obj)
    good('collection[*]', obj)
    good('collection[*](*)', obj)
    good('collection[N]', obj)
    fail('collection[N],N>0', [])

for obj in ([1.], (1.,), {1.}, frozenset((1.,)), collections.deque((1.,)), {1.:1},):
    good('collection[*](float)', obj)
    good('collection[=1]', obj)
    fail('collection[=2]', obj)
    good('collection[1]', obj)  # shortcut

    good('collection[N],N>0', obj)
    good('collection[N],N=1', obj)
    good('collection[N],N>0,N<2', obj)

    fail('collection[*](int)', obj)




# cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
# Aspidites
# Copyright (C) 2021 Ross J. Duff

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import itertools
from collections import OrderedDict
from typing import Union, ItemsView
from textwrap import wrap
from Aspidites._vendor.pyparsing import ParseResults
from inspect import isfunction, signature
from ._vendor.contracts import new_contract
from ._vendor.fn.underscore import ArityError, _Callable
from .templates import warning_template
import cython


def _wrap_lines(text, padchar, width, wrapped_lines, pad):
    for l in text.splitlines():
        line = wrap(l, width, replace_whitespace=False)
        for s in line:
            s += padchar * width * pad
        wrapped_lines.extend(line)
    return wrapped_lines


def _wrap(text, width, pad, padchar):
    """
    Do not remove whitespaces in string but still wrap text to max width.
    Instead of passing the entire text to textwrap.wrap, split and pass each
    line instead. This way list formatting is not mangled by textwrap.wrap.
    """
    pad = pad or True
    padchar = padchar or ' '.encode('UTF-8')
    width = width or 160

    wrapped_lines = []
    try:
        _wrap_lines(text, padchar, width, wrapped_lines, pad)
    except ArityError:
        wrapped_lines.extend([str(text)])

    return wrapped_lines


def bordered(text, width=160) -> str:
    """Create a fancy bordered textbox as used in :class:`Aspidites.api.Warn`."""
    i: str
    lines: list = [t for t in _wrap(text, width=width, pad=True, padchar=" ")]
    lens: list = [len(i) for i in lines]
    width = max(lens, default=width) or width
    res = ["╭" + "┉" * width + "╮"]
    for s in lines:
        while len(s) < width:
            s += " "
        res.append("┊" + s + "┊")
    res.append("╰" + "┉" * width + "╯")
    return "\n".join(res)


def _format_locals(lokals: dict):
    d = {k: v for k, v in lokals.items() if not str(k).startswith('@')}
    str_locals = _format_items(d)
    return str_locals


def _format_items(x):
    str_locals = ''.encode('UTF-8')
    for k, v in x.items():
        if isfunction(v):
            s = str(signature(v)).replace("'", "")
            str_locals += f"{k}: {s}\n".encode('UTF-8')
        else:
            str_locals += f"{k}: {str(v)}\n".encode('UTF-8')
    return str_locals


class Warn:
    """Creates a nice human-readable warning with a fancy border."""
    sig: cython.p_char
    name: Union[_Callable, cython.p_char]
    local_items: ItemsView
    str_locals: bytes
    at_fault: bytes
    func_name: cython.p_char
    fname: cython.p_char
    # noinspection PyUnresolvedReferences
    lineno: cython.int
    fkwargs: cython.p_char

    def __init__(self, stack, func, *args, **kwargs):
        self.stack = stack
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def create(self, exc=Exception()) -> str:
        local_items = self.stack[1][0].f_locals.items()
        str_locals = self.format_locals(local_items, exc)
        func_name = self.stack[1][0].f_code.co_name
        fname = self.stack[1][0].f_code.co_filename
        lineno = self.stack[1][0].f_code.co_firstlineno
        fkwargs = str(self.format_kwargs()).lstrip("b'").rstrip("'")
        name = self.func.__name__ if hasattr(self.func, "__name__") else str(self.func)
        sig = f"{name}({str(self.args).strip('()')}{fkwargs})"
        at_fault = str(name).encode('UTF-8') if isinstance(exc, TypeError) else sig.encode('UTF-8')
        return warning_template.safe_substitute(
            file=fname,
            lineno=lineno,
            func=bordered(func_name),
            atfault=bordered(at_fault.decode('UTF-8')),
            bound=bordered(str_locals.decode('UTF-8')),
            tb=bordered(str(exc)),
        )

    def format_kwargs(self, sep: str = ", ") -> bytes:
        return f'{sep}{str(self.kwargs).strip("{} ").replace(":", "=") if len(self.kwargs) else ""}'.encode('UTF-8')

        # noinspection PyMethodMayBeStatic

    def format_locals(self, local_vars, exc: Exception) -> bytes:
        OrderedDict(itertools.islice(local_vars, 10))
        lokals: dict = dict(filter(lambda x: x[1] != str(exc), local_vars))
        str_locals: bytes = _format_locals(lokals)
        return str_locals.rstrip("\n".encode('UTF-8'))


code = new_contract("code", lambda x: isinstance(x, ParseResults))



from Aspidites._vendor._compat import IS_PY38_OR_GREATER
from Aspidites.tests._vendor.RestrictedPython.helper import restricted_exec


DICT_COMPREHENSION_WITH_ATTRS = """
def call(seq):
    return {y.k: y.v for y in seq.z if y.k}
"""


def test_dict_comprehension_with_attrs(mocker):
    _getattr_ = mocker.Mock()
    _getattr_.side_effect = getattr

    _getiter_ = mocker.Mock()
    _getiter_.side_effect = lambda ob: ob

    glb = {'_getattr_': _getattr_, '_getiter_': _getiter_}
    restricted_exec(DICT_COMPREHENSION_WITH_ATTRS, glb)

    z = [mocker.Mock(k=0, v='a'), mocker.Mock(k=1, v='b')]
    seq = mocker.Mock(z=z)

    ret = glb['call'](seq)
    assert ret == {1: 'b'}

    _getiter_.assert_called_once_with(z)

    calls = [mocker.call(seq, 'z')]

    # Note: Order changed in PEP 572, starting with Python 3.8.
    if IS_PY38_OR_GREATER:
        calls.extend([
            mocker.call(z[0], 'k'),
            mocker.call(z[1], 'k'),
            mocker.call(z[1], 'k'),
            mocker.call(z[1], 'v'),
        ])
    else:
        calls.extend([
            mocker.call(z[0], 'k'),
            mocker.call(z[1], 'k'),
            mocker.call(z[1], 'v'),
            mocker.call(z[1], 'k'),
        ])

    _getattr_.assert_has_calls(calls)



""" Test for bibparse grammar """

import unittest
from Aspidites._vendor.pyparsing import ParseException
from .btpyparse import Macro
from . import btpyparse as bp


class TestBibparse(unittest.TestCase):
    def test_names(self):
        # check various types of names
        # All names can contains alphas, but not some special chars
        bad_chars = '"#%\'(),={}'
        for name_type, dig1f in ((bp.macro_def, False),
                                 (bp.field_name, False),
                                 (bp.entry_type, False),
                                 (bp.cite_key, True)):
            if dig1f:  # can start with digit
                self.assertEqual(name_type.parseString('2t')[0], '2t')
            else:
                self.assertRaises(ParseException, name_type.parseString, '2t')
            # All of the names cannot contain some characters
            for char in bad_chars:
                self.assertRaises(ParseException, name_type.parseString, char)
            # standard strings all OK
            self.assertEqual(name_type.parseString('simple_test')[0], 'simple_test')
        # Test macro ref
        mr = bp.macro_ref
        # can't start with digit
        self.assertRaises(ParseException, mr.parseString, '2t')
        for char in bad_chars:
            self.assertRaises(ParseException, mr.parseString, char)
        self.assertEqual(mr.parseString('simple_test')[0].name, 'simple_test')

    def test_numbers(self):
        self.assertEqual(bp.number.parseString('1066')[0], '1066')
        self.assertEqual(bp.number.parseString('0')[0], '0')
        self.assertRaises(ParseException, bp.number.parseString, '-4')
        self.assertRaises(ParseException, bp.number.parseString, '+4')
        self.assertRaises(ParseException, bp.number.parseString, '.4')
        # something point something leaves a trailing .4 unmatched
        self.assertEqual(bp.number.parseString('0.4')[0], '0')

    def test_parse_string(self):
        # test string building blocks
        self.assertEqual(bp.chars_no_quotecurly.parseString('x')[0], 'x')
        self.assertEqual(bp.chars_no_quotecurly.parseString("a string")[0], 'a string')
        self.assertEqual(bp.chars_no_quotecurly.parseString('a "string')[0], 'a ')
        self.assertEqual(bp.chars_no_curly.parseString('x')[0], 'x')
        self.assertEqual(bp.chars_no_curly.parseString("a string")[0], 'a string')
        self.assertEqual(bp.chars_no_curly.parseString('a {string')[0], 'a ')
        self.assertEqual(bp.chars_no_curly.parseString('a }string')[0], 'a ')
        # test more general strings together
        for obj in (bp.curly_string, bp.string, bp.field_value):
            self.assertEqual(obj.parseString('{}').asList(), [])
            self.assertEqual(obj.parseString('{a "string}')[0], 'a "string')
            self.assertEqual(obj.parseString('{a {nested} string}').asList(),
                             ['a ', ['nested'], ' string'])
            self.assertEqual(obj.parseString('{a {double {nested}} string}').asList(),
                             ['a ', ['double ', ['nested']], ' string'])
        for obj in (bp.quoted_string, bp.string, bp.field_value):
            self.assertEqual(obj.parseString('""').asList(), [])
            self.assertEqual(obj.parseString('"a string"')[0], 'a string')
            self.assertEqual(obj.parseString('"a {nested} string"').asList(),
                             ['a ', ['nested'], ' string'])
            self.assertEqual(obj.parseString('"a {double {nested}} string"').asList(),
                             ['a ', ['double ', ['nested']], ' string'])
        # check macro def in string
        self.assertEqual(bp.string.parseString('someascii')[0], Macro('someascii'))
        self.assertRaises(ParseException, bp.string.parseString, '%#= validstring')
        # check number in string
        self.assertEqual(bp.string.parseString('1994')[0], '1994')

    def test_parse_field(self):
        # test field value - hashes included
        fv = bp.field_value
        # Macro
        self.assertEqual(fv.parseString('aname')[0], Macro('aname'))
        self.assertEqual(fv.parseString('ANAME')[0], Macro('aname'))
        # String and macro
        self.assertEqual(fv.parseString('aname # "some string"').asList(),
                         [Macro('aname'), 'some string'])
        # Nested string
        self.assertEqual(fv.parseString('aname # {some {string}}').asList(),
                         [Macro('aname'), 'some ', ['string']])
        # String and number
        self.assertEqual(fv.parseString('"a string" # 1994').asList(),
                         ['a string', '1994'])
        # String and number and macro
        self.assertEqual(fv.parseString('"a string" # 1994 # a_macro').asList(),
                         ['a string', '1994', Macro('a_macro')])

    def test_comments(self):
        res = bp.comment.parseString('@Comment{about something}')
        self.assertEqual(res.asList(), ['comment', '{about something}'])
        self.assertEqual(
            bp.comment.parseString('@COMMENT{about something').asList(),
            ['comment', '{about something'])
        self.assertEqual(
            bp.comment.parseString('@comment(about something').asList(),
            ['comment', '(about something'])
        self.assertEqual(
            bp.comment.parseString('@COMment about something').asList(),
            ['comment', ' about something'])
        self.assertRaises(ParseException, bp.comment.parseString,
                          '@commentabout something')
        self.assertRaises(ParseException, bp.comment.parseString,
                          '@comment+about something')
        self.assertRaises(ParseException, bp.comment.parseString,
                          '@comment"about something')

    def test_preamble(self):
        res = bp.preamble.parseString('@preamble{"about something"}')
        self.assertEqual(res.asList(), ['preamble', 'about something'])
        self.assertEqual(bp.preamble.parseString(
            '@PREamble{{about something}}').asList(),
                         ['preamble', 'about something'])
        self.assertEqual(bp.preamble.parseString("""@PREamble{
            {about something}
        }""").asList(),
                         ['preamble', 'about something'])

    def test_macro(self):
        res = bp.macro.parseString('@string{ANAME = "about something"}')
        self.assertEqual(res.asList(), ['string', 'aname', 'about something'])
        self.assertEqual(
            bp.macro.parseString('@string{aname = {about something}}').asList(),
            ['string', 'aname', 'about something'])

    def test_entry(self):
        txt = """@some_entry{akey, aname = "about something",
        another={something else}}"""
        res = bp.entry.parseString(txt)
        self.assertEqual(res.asList(),
                         ['some_entry', 'akey',
                          ['aname', 'about something'], ['another', 'something else']])
        # Case conversion
        txt = """@SOME_ENTRY{akey, ANAME = "about something",
        another={something else}}"""
        res = bp.entry.parseString(txt)
        self.assertEqual(res.asList(),
                         ['some_entry', 'akey',
                          ['aname', 'about something'], ['another', 'something else']])

    def test_bibfile(self):
        txt = """@some_entry{akey, aname = "about something",
        another={something else}}"""
        res = bp.bibfile.parseString(txt)
        self.assertEqual(res.asList(),
                         [['some_entry', 'akey',
                           ['aname', 'about something'],
                           ['another', 'something else']]])

    def test_bib1(self):
        # First pass whole bib-like tests
        txt = """
    Some introductory text
    (implicit comment)

        @ARTICLE{Brett2002marsbar,
      author = {Matthew Brett and Jean-Luc Anton and Romain Valabregue and Jean-Baptise
                Poline},
      title = {{Region of interest analysis using an SPM toolbox}},
      journal = {Neuroimage},
      year = {2002},
      volume = {16},
      pages = {1140--1141},
      number = {2}
    }

    @some_entry{akey, aname = "about something",
    another={something else}}
    """
        res = bp.bibfile.parseString(txt)
        self.assertEqual(len(res), 3)
        res2 = bp.parse_str(txt)
        self.assertEqual(res.asList(), res2.asList())
        res3 = [r.asList()[0] for r, start, end in bp.definitions.scanString(txt)]
        self.assertEqual(res.asList(), res3)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class ExceptionsTest(unittest.TestCase):

    def test_match_error_in_terse_syntax(self):
        with self.assertRaises(MatchError):
            match(7,
                  0, "zero",
                  1, "one")

    def test_match_error_in_declarative_syntax(self):
        @case_distinction
        def f(x: Match(0)):
            pass

        @case_distinction
        def f(x: Match(1)):
            pass

        with self.assertRaises(MatchError):
            f(2)



#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from Aspidites._vendor._compat import basestring

from ..interface import Contract, ContractNotRespected, RValue
from ..syntax import W
from ..interface import eval_in_context


class EqualTo(Contract):
    def __init__(self, rvalue, where=None):
        Contract.__init__(self, where)
        assert isinstance(rvalue, RValue), "Expected RValue for EqualTo, got %r" % rvalue
        self.rvalue = rvalue

    def check_contract(self, context, value, silent):
        val = eval_in_context(context, self.rvalue, self)
        if not(val == value):
            error = ('EqualTo: Condition %s == %s not respected.'
                     % (val, value))
            raise ContractNotRespected(contract=self, error=error,
                                       value=value, context=context)

    def __str__(self):
        return "%s" % self.rvalue

    def __repr__(self):
        return 'EqualTo(%r)' % self.rvalue

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        rvalue = tokens[0]
        from Aspidites._vendor.contracts.library.types_misc import CheckType
        if isinstance(rvalue, CheckType):
            return rvalue
        else:
            assert isinstance(rvalue, RValue)
            return EqualTo(rvalue, where)


class SimpleRValue(RValue):
    def __init__(self, value, representation=None, where=None):
        assert representation is None or isinstance(representation, basestring)
        self.value = value
        self.where = where
        self.representation = representation

    def __str__(self):
        if self.representation is None:
            return "{0!s}".format(self.value)
        else:
            return self.representation

    def __repr__(self):
        if self.representation is None:
            return "SimpleRValue({0!r})".format(self.value)
        else:
            return "SimpleRValue({0!r},{1!r})".format(self.value,
                                                      self.representation)

    def eval(self, context):  # @UnusedVariable @ReservedAssignment
        return self.value





#cython: language_level=3, annotation_typing=True, c_string_encoding=utf-8, boundscheck=False, wraparound=False, initializedcheck=False
from ..interface import Contract, ContractNotRespected
from ..syntax import (add_contract, W, contract_expression, O, S, add_keyword,
    Keyword)
import collections.abc as collections
from past.builtins import xrange

try:
    import numpy
    has_numpy = True
except ImportError:
    has_numpy = False


class Seq(Contract):

    def __init__(self, length_contract=None,
                    elements_contract=None, where=None):
        Contract.__init__(self, where)
        self.length_contract = length_contract
        self.elements_contract = elements_contract

    def check_contract(self, context, value, silent):
        if has_numpy and isinstance(value, numpy.ndarray):
            # use value.size and value.flat for iteration
            if self.length_contract is not None:
                self.length_contract._check_contract(context, value.size, silent)

            if self.elements_contract is not None:
                n = value.size
                for i in xrange(n):
                    element = value.flat[i]
                    if ((element.dtype == numpy.int32) or
                        (element.dtype == numpy.int64)):
                        element = int(element)
                    # XXX: hack
                    self.elements_contract._check_contract(context, element, silent)

            return

        if not isinstance(value, collections.Sequence):
            error = 'Expected a Sequence, got %r.' % value.__class__.__name__
            raise ContractNotRespected(self, error, value, context)

        if self.length_contract is not None:
            self.length_contract._check_contract(context, len(value), silent)

        if self.elements_contract is not None:
            for element in value:
                self.elements_contract._check_contract(context, element, silent)

    def __str__(self):
        s = 'seq'
        if self.length_contract is not None:
            s += '[%s]' % self.length_contract
        if self.elements_contract is not None:
            s += '(%s)' % self.elements_contract
        return s

    def __repr__(self):
        s = 'Seq(%r,%r)' % (self.length_contract, self.elements_contract)
        return s

    @staticmethod
    def parse_action(s, loc, tokens):
        where = W(s, loc)
        length_contract = tokens.get('length_contract', None)
        elements_contract = tokens.get('elements_contract', None)
        return Seq(length_contract, elements_contract, where=where)


list_contract = (Keyword('seq') -
                 O(S('[') - contract_expression('length_contract') - S(']')) +
                 O(S('(') - contract_expression('elements_contract') - S(')')))
list_contract.setParseAction(Seq.parse_action)

list_contract.setName('Seq contract')
add_keyword('seq')
add_contract(list_contract)



#
# list1.py
#
# an example of using parse actions to convert type of parsed data.
#
# Copyright (c) 2006-2016, Paul McGuire
#
from pyparsing import *

# first pass
lbrack = Literal("[")
rbrack = Literal("]")
integer = Word(nums).setName("integer")
real = Combine(Optional(oneOf("+ -")) + Word(nums) + "." +
               Optional(Word(nums))).setName("real")

listItem = real | integer | quotedString

listStr = lbrack + delimitedList(listItem) + rbrack

test = "['a', 100, 3.14]"

print(listStr.parseString(test))


# second pass, cleanup and add converters
lbrack = Literal("[").suppress()
rbrack = Literal("]").suppress()
cvtInt = lambda s,l,toks: int(toks[0])
cvtReal = lambda s,l,toks: float(toks[0])
integer = Word(nums).setName("integer").setParseAction( cvtInt )
real = Combine(Optional(oneOf("+ -")) + Word(nums) + "." +
               Optional(Word(nums))).setName("real").setParseAction( cvtReal )
listItem = real | integer | quotedString.setParseAction( removeQuotes )

listStr = lbrack + delimitedList(listItem) + rbrack

test = "['a', 100, 3.14]"

print(listStr.parseString(test))

# third pass, add nested list support
lbrack, rbrack = map(Suppress, "[]")

cvtInt = tokenMap(int)
cvtReal = tokenMap(float)

integer = Word(nums).setName("integer").setParseAction( cvtInt )
real = Regex(r"[+-]?\d+\.\d*").setName("real").setParseAction( cvtReal )

listStr = Forward()
listItem = real | integer | quotedString.setParseAction(removeQuotes) | Group(listStr)
listStr << lbrack + delimitedList(listItem) + rbrack

test = "['a', 100, 3.14, [ +2.718, 'xyzzy', -1.414] ]"
print(listStr.parseString(test))


from __future__ import annotations

import unittest

from Aspidites._vendor.apm import *


class CaseOfTest(unittest.TestCase):

    # noinspection PyShadowingBuiltins
    def test_case_of(self):
        result = case({"Id": "h23ksad8", "Name": "ObjectName"}) \
            .of({"Id": Capture(Regex("[A-Z0-9]+"), name='id')}, lambda id: id) \
            .of({"Id": Capture(Regex("[a-z0-9]+"), name='id')}, lambda id: id.upper()) \
            .otherwise("_")
        self.assertEqual("H23KSAD8", result)

    # noinspection PyShadowingBuiltins
    def test_case_of_constant(self):
        result = case({"Id": "h23ksad8", "Name": "ObjectName"}) \
            .of({"Id": Capture(Regex("[A-Z0-9]+"), name='id')}, "upper") \
            .of({"Id": Capture(Regex("[a-z0-9]+"), name='id')}, "lower") \
            .otherwise("non")
        self.assertEqual("lower", result)

    # noinspection PyShadowingBuiltins
    def test_case_of_default_action(self):
        result = case({"Id": "23892", "Name": "ObjectName"}) \
            .of({"Id": Regex("[A-Z][a-z0-9]*")}, "upper") \
            .of({"Id": Regex("[a-z][a-z0-9]*")}, "lower") \
            .otherwise(lambda: "something else")
        self.assertEqual("something else", result)

    def test_case_of_default(self):
        result = case({"Id": "h23ksad8", "Name": "ObjectName"}) \
            .of({"Id": Capture(Regex("[0-9]+"), name='id')}, "numeric") \
            .otherwise("non-numeric")
        self.assertEqual("non-numeric", result)



from Aspidites._vendor.contracts import new_contract
from Aspidites._vendor.contracts.test_registrar import fail, good, syntax_fail



@new_contract
def ext0_positive(value):
    return value > 0

@new_contract
def ext1_lessthan(value, threshold):
    return value < threshold


good('ext0_positive', 1)
fail('ext0_positive', -1)


good('ext1_lessthan(0)', -1)  
fail('ext1_lessthan(0)', +1)
# named
good('ext1_lessthan(threshold=0)', -1)  
fail('ext1_lessthan(threshold=0)', +1)


# needs to fail parsing because we didn't provide argument
syntax_fail('ext1_lessthan')
# needs to fail parsing because the argument name is wrong
syntax_fail('ext1_lessthan(th=0)')
# too many arguments
syntax_fail('ext1_lessthan(0,1)')




from __future__ import annotations

import unittest
# noinspection PyProtectedMember
from dataclasses import dataclass

from Aspidites._vendor.apm import *


@dataclass
class Foo:
    x: int


class UtilTests(unittest.TestCase):

    def test_issue_2(self):
        """https://github.com/scravy/awesome-pattern-matching/issues/2"""

        result = match(Foo(1), Foo, 'this matches', 'no match unfortunately')
        self.assertEqual('this matches', result)

        self.assertTrue(match(Foo(1), Foo, True, False))
        self.assertTrue(match(Foo(1), Foo(1), True, False))
        self.assertFalse(match(Foo(1), Foo(2), True, False))



from __future__ import annotations

import unittest
from typing import Callable

from Aspidites._vendor.apm import *


class AggregationsTest(unittest.TestCase):

    def test_agg_set(self):
        result = match([
            'foo',
            3,
            'bar',
            10,
            -4,
            'foo',
            'bar',
            17,
        ], Each(OneOf(InstanceOf(str) >> agg.Set('strings'), ...)))
        self.assertTrue(result)
        self.assertEqual(result.strings, {'foo', 'bar'})

    def test_agg_list(self):
        result = match([
            'foo',
            3,
            'bar',
            10,
            -4,
            'foo',
            'bar',
            17,
        ], Each(OneOf(InstanceOf(str) >> agg.List('strings'), ...)))
        self.assertTrue(result)
        self.assertEqual(result.strings, ['foo', 'bar', 'foo', 'bar'])

    def test_agg_count(self):
        result = match([
            1,
            2,
            3.0,
            4,
            5.0,
        ], Each(OneOf(
            InstanceOf(int) >> agg.Count('integers'),
            InstanceOf(float) >> agg.Count('floats'),
        )))
        self.assertTrue(result)
        self.assertEqual(3, result.integers)
        self.assertEqual(2, result.floats)

    def test_agg_histogram(self):
        result = match([
            'a', 'b', 'c', 'b', 'd', 'c', 'a', 'a', 'c', 'a'
        ], Each(_ >> agg.Histogram('histo')))
        self.assertTrue(result)
        self.assertEqual({
            'a': 4,
            'b': 2,
            'c': 3,
            'd': 1,
        }, result.histo)

    def test_agg_sum(self):
        lower = 1
        upper = 10
        result = match(range(lower, upper), Each(_ >> agg.Sum('sum')))
        length = upper - lower
        self.assertTrue(result)
        self.assertEqual(length * (length + 1) // 2, result.sum)

    def test_agg_float_sum(self):
        result = match(['1', '10.3'], Each(_ >> agg.Sum('sum')))
        self.assertTrue(result)
        self.assertEqual(11.3, result.sum)

    def test_some_more_complicated_pattern(self):
        class Recursive(Pattern):
            def __init__(self, p: Callable):
                self._pattern = p

            def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
                return ctx.match(value, self._pattern())

        numbers = agg.Set()
        pattern = OneOf(InstanceOf(int) >> numbers, EachItem(..., Recursive(lambda: pattern)))

        result = match({
            'a': {'b': 3},
            'c': 7,
            'd': {'e': {}},
            'f': {'g': {'h': 10, 'i': 7}},
        }, pattern)
        self.assertTrue(result)
        self.assertEqual({3, 7, 10}, numbers.value)

    def test_some_more_complicated_pattern_with_nothing_in_it(self):
        class Recursive(Pattern):
            def __init__(self, p: Callable):
                self._pattern = p

            def match(self, value, *, ctx: MatchContext, strict: bool) -> MatchResult:
                return ctx.match(value, self._pattern())

        numbers = agg.Set()
        pattern = OneOf(InstanceOf(float) >> numbers, InstanceOf(int), EachItem(..., Recursive(lambda: pattern)))

        result = match({
            'a': {'b': 3},
            'c': 7,
            'd': {'e': {}},
            'f': {'g': {'h': 10, 'i': 7}},
        }, pattern)
        self.assertTrue(result)
        self.assertEqual(set(), numbers.value)



from . import agg
from .__pkginfo__ import __version__
from .case_of import case
from .core import \
    AllOf, \
    Capture, \
    Either, \
    MatchContext, \
    MatchResult, \
    Not, \
    OneOf, \
    Pattern, \
    Many, \
    Strict, \
    String, \
    StringPattern, \
    Value, \
    _
from .error import MatchError
from .match import match
from .overload import case_distinction, Match
from .patterns import \
    Arguments, \
    At, \
    Attrs, \
    Between, \
    Check, \
    Contains, \
    Each, \
    EachItem, \
    InstanceOf, \
    IsNumber, \
    IsString, \
    IsTruish, \
    Length, \
    Maybe, \
    NoneOf, \
    Items, \
    Regex, \
    Returns, \
    SubclassOf, \
    Transformed
from .try_match import Case, Default
from .typefoo import \
    Parameters, \
    KwArgs, \
    VarArgs

Remaining = Many
Some = Many

Object = Items

__all__ = [
    'agg',

    'AllOf',
    'Capture',
    'Either',
    'Many',
    'MatchContext',
    'MatchResult',
    'Not',
    'OneOf',
    'Pattern',
    'Remaining',
    'Some',
    'Strict',
    'String',
    'StringPattern',
    'Value',

    '_',

    'case',
    'match',
    'case_distinction',
    'Match',
    'MatchError',
    'Case',
    'Default',

    'Arguments',
    'At',
    'Attrs',
    'Between',
    'Check',
    'Contains',
    'Each',
    'EachItem',
    'InstanceOf',
    'IsNumber',
    'IsString',
    'IsTruish',
    'Items',
    'Length',
    'Maybe',
    'NoneOf',
    'Object',
    'Regex',
    'Returns',
    'SubclassOf',
    'Transformed',

    'Parameters',
    'KwArgs',
    'VarArgs',
]
